/*
 * Tencent is pleased to support the open source community by making  XiaoweiSDK Demo Codes available.
 *
 * Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
 *
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://opensource.org/licenses/MIT
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 *
 */
package com.tencent.xiaowei.sdk;
import android.util.Log;
import com.tencent.xiaowei.info.DeviceBinderInfo;
import com.tencent.xiaowei.info.DeviceLoginInfo;
import java.net.URI;
import java.security.MessageDigest;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * SDK基础功能
 */
public class DeviceBaseManager {
    public static final String TAG = "DeviceBaseManager";
    // 绑定者相关
    private static Runnable refreshQrCodeRunnable = null;
    protected final static HashMap<String, DeviceBinderInfo> mAllFriendListCache = new HashMap<>();
    private static OnGetBinderListListener mOnGetBinderListListener;
    private static OnGetSDKLogListener mOnGetSDKLogListener;
    private static DeviceSDK.OnDeviceLoginListener onDeviceLoginListener;
    private static OnTicketRefreshListener mOnTicketRefreshListener;
    static int qrCodeRetry = 0;

    /**
     * 用于蓝牙或声波配网的ticket刷新，测试使用
     */
    public interface OnTicketRefreshListener {
        /**
         * 获取到的ticket
         * @param ticket
         * @param expireTime
         * @param expireIn
         */
        void onRefresh(String ticket, int expireTime, int expireIn);
    }

    /**
     * 主动获取用于声波配网的ticket，并设置回调
     * @param listener
     */
    public static void getTicket(OnTicketRefreshListener listener) {
        if(listener != null) {
            mOnTicketRefreshListener = listener;
            Log.d(TAG, "getTicket");
            XWSDKJNI.getBindTicket();
        }
    }

    /**
     * JNI层上来的ticket获取结果
     */
    static void onTicketRefresh(final String ticket , final int expireTime, final int expireIn) {
        if (mOnTicketRefreshListener != null) {
            mOnTicketRefreshListener.onRefresh(ticket, expireTime, expireIn);
        }
    }

    /**
     * 设置设备基础事件listener，一般来说必须要设置，否者SDK难以正常工作
     * @param listener
     */
    static void setOnDeviceLoginListener(DeviceSDK.OnDeviceLoginListener listener) {
        DeviceBaseManager.onDeviceLoginListener = listener;
    }

    /**
     * 主动获取绑定者列表的结果，被动获取的在onDeviceLoginListener里
     */
    public interface OnGetBinderListListener {
        void onResult(int error, ArrayList<DeviceBinderInfo> mBinderList);
    }

    /**
     * 主动获取绑定者列表，设置一个listener来获得结果，一般无需设置
     *
     * @param listener
     */
    public static void getBinderList(OnGetBinderListListener listener) {
        DeviceBinderInfo[] list = XWSDKJNI.fetchBinderList();
        if (list != null && listener != null) {
            listener.onResult(0, XWSDKJNI.getBinderList());
        } else {
            mOnGetBinderListListener = listener;
        }
    }

    /**
     * 设备绑定者产生变化的通知，每当绑定者任何信息发生变化时，都会全量回调最新的所有绑定者信息
     * @param error 一般无需关注
     * @param mBinderList 最新的绑定者列表
     */
    static void onBinderListChange(final int error, final ArrayList<DeviceBinderInfo> mBinderList) {
        postMain(new Runnable() {
            @Override
            public void run() {
                if (mOnGetBinderListListener != null) {
                    mOnGetBinderListListener.onResult(error, mBinderList);
                    mOnGetBinderListListener = null;
                } else {
                    if (onDeviceLoginListener != null) {
                        onDeviceLoginListener.onBinderListChange(error, mBinderList);
                    }
                }
            }
        });
        Log.d(TAG, "onBinderListChange " + error);
    }

    /**
     * 好友绑定请求，有人扫了二维码，发送请求过来，注意这个回调是用户扫码之后，点击绑定的时候才会触发。可以在设备端同意或拒绝
     * @param info
     */
    static void onBinderVerify(final DeviceBinderInfo info) {
        if (onDeviceLoginListener != null) {
            postMain(new Runnable() {
                @Override
                public void run() {
                    onDeviceLoginListener.onBinderVerify(info);
                }
            });
        }
        Log.d(TAG, "onBinderVerify " + new String(info.nickName));
    }

    /**
     * 二维码刷新的通知，无论是主动拉二维码，还是二维码被使用了，都会通过这个接口回调。
     * 本质上在二维码过期后SDK不会自动刷新它，但在这个接口里面，顺便实现了一下自动刷新逻辑，可以参考下面的代码
     * @param path 二维码本地路径，如果是空则表示获取失败
     * @param expireTime 二维码过期相对时间，单位秒
     * @param expireIn 二维码过期绝对时间，单位秒
     */
    static void onQrCodeRefresh(final String path , final int expireTime, final int expireIn) {
        if (onDeviceLoginListener != null) {
            postMain(new Runnable() {
                @Override
                public void run() {
                    onDeviceLoginListener.onQrCodeRefresh(path, expireTime, expireIn);
                }
            });
        }
        Log.d(TAG, "onQrCodeRefresh, expireTime =  " + expireTime + ", expire in "+ expireIn);
        int timeDelay = 0;
        // 如果获取二维码失败了，5秒后自动重试
        if(expireTime == 0 || path.length() == 0){
            timeDelay = 5;
            qrCodeRetry++;
        }
        else {
            // 90%时间重试一把
            timeDelay = (int)(expireIn * 0.9);
            // fix一下重试时间，过长或过短都fix一下
            if (timeDelay < 60 || timeDelay > 120*60 ) {
                timeDelay = 60 * 30;
            }
            qrCodeRetry = 0;
        }
        // 最多自动重试10次
        if (qrCodeRetry > 10) {
            Log.e(TAG, "qrCodeRetry too much times");
            return;
        }
        
        Log.d(TAG, "onQrCodeRefresh, next auto refresh time delay  =  " + timeDelay + ",current time = " +System.currentTimeMillis()/1000);
        // 每次执行到这里，先取消之前的自动刷新
        XWSDKJNI.removeMainRunnable(refreshQrCodeRunnable);
        refreshQrCodeRunnable = new Runnable() {
            @Override
            public void run() {
                DeviceBaseManager.refreshQRCodeUrl();
            }
        };
        XWSDKJNI.postMainDelay(refreshQrCodeRunnable,timeDelay *1000);
    }




    /**
     * 解绑所有绑定者（须在登录成功之后调用)，解绑成功后正常就会收到绑定者变化的回调了，这里就不需要回调了
     */
    public static void unBindAll() {
        XWSDKJNI.eraseAllBinders();
        Log.d(TAG, "eraseAllBinders ");
    }

    /**
     * 重置设备，设备所有信息将被清空，慎用
     * @return 0成功，其它失败
     */
   public static int factoryReset() {
       Log.e(TAG, "factoryReset");
       return XWSDKJNI.factoryReset();
   }

    /**
     * 解绑某一个绑定者（须在登录成功之后调用)，解绑成功后正常就会收到绑定者变化的回调了，这里就不需要回调了
      * @param userName 绑定者的userName
     */
    public static void unBindByUserName(String userName) {
        XWSDKJNI.eraseBinder(userName);
        Log.d(TAG, "eraseBinder by username" + userName);
    }

    /**
     * 获取当前所有的好友请求，一般不需要，主要是针对加好友的时候设备不在线，设备恢复上线的时候可以拉一下。
     * @return
     */
    public static DeviceBinderInfo[] getUnverifyBinderList() {
        Log.d(TAG, "getUnverifyBinderList");
        return XWSDKJNI.getUnverifyBinderList();
    }

    /**
     * 处理好友请求验证
     *
     * @param binderUserName 要处理的用户名
     * @param isAccept true同意，false拒绝
     * @return 成功或者失败
     */
    public static int handleBinderVerify(String binderUserName, boolean isAccept)  {
        Log.d(TAG, "handleBinderVerify");
        return  XWSDKJNI.handleBinderVerify(binderUserName,isAccept);
    }

    /**
     * 改好友别名
     * 
     * @param binderUserName 要处理的用户名
     * @param alias 新的备注 
     * @return 成功或者失败
     */
    public static int updateBinderAlias(String binderUserName, String alias)  {
        Log.d(TAG, "updateBinderAlias");
        return XWSDKJNI.updateBinderAlias(binderUserName, alias);
    }

    // 基础接口

    /**
     * 获得SDK版本
     *
     * @return
     */
    public static int[] getSDKVersion() {
        return XWSDKJNI.getInstance().getDeviceSDKVersion();
    }


    /**
     * 主动刷新二维码url
     *
     * @return
     */
    public static int refreshQRCodeUrl() {
        return XWSDKJNI.refreshQRCodeUrl();
    }


    /**
     * 获取Uin
     */
    public static long getUin() {
        return XWSDKJNI.getUin();
    }

    /**
     * 获取deviceId
     */
    public static String getDeviceId() {
        return XWSDKJNI.getDeviceId();
    }

    public static String MD5(String path) {
        byte[] buffer = path.getBytes();
        try {
            MessageDigest md5 = MessageDigest.getInstance("MD5");
            byte[] md5Bytes = md5.digest(buffer);
            StringBuffer hexValue = new StringBuffer();
            for (int i = 0; i < md5Bytes.length; i++) {
                int val = ((int) md5Bytes[i]) & 0xff;
                if (val < 16)
                    hexValue.append("0");
                hexValue.append(Integer.toHexString(val));
            }
            return hexValue.toString();
        } catch (Exception e) {
            e.printStackTrace();
        }
        return null;
    }


    public static String appendUri(String uri, String appendQuery) {

        try {
            URI oldUri = new URI(uri);

            String newQuery = oldUri.getQuery();
            if (newQuery == null) {
                newQuery = appendQuery;
            } else {
                newQuery += "&" + appendQuery;
            }

            URI newUri = new URI(oldUri.getScheme(), oldUri.getAuthority(),
                    oldUri.getPath(), newQuery, oldUri.getFragment());

            return newUri.toString();

        } catch (Exception e) {
            e.printStackTrace();
        }

        return uri;
    }

    /**
     * 获取SDK的native日志，受login时候传入的参数日志级别的影响
     */
    public interface OnGetSDKLogListener {
        /**
         * 厂商日志获取
         *
         * @param level   log级别
         * @param filename  文件名
         * @param line    行号
         * @param funcname 函数名
         * @param message log内容
         */
        void onGetSDKLog(int level, String tag, String filename, int line, String funcname, String message);
    }

    /**
     * 设置日志回调，如果设置了，那么SDK底层日志会通脱它回调到上层
     * @param listener
     */
    public static void setOnGetSDKLogListener(OnGetSDKLogListener listener) {
        mOnGetSDKLogListener = listener;
    }

    static void onGetSDKLog(int level, String tag, String filename, int line, String funcname, String message) {
        if (mOnGetSDKLogListener != null) {
            mOnGetSDKLogListener.onGetSDKLog(level, tag, filename, line, funcname, message);
        }
    }

    /** 日志end */

    /** OnDeviceLoginListener begin */
    static void onLoginComplete(final int error) {
        if (onDeviceLoginListener != null) {
            postMain(new Runnable() {
                @Override
                public void run() {
                    onDeviceLoginListener.onLoginComplete(error);
                }
            });
        }
    }

    static void onNetStatusChange(final int newStatus) {
        if (onDeviceLoginListener != null) {
            postMain(new Runnable() {
                @Override
                public void run() {
                    onDeviceLoginListener.onNetStatusChange(newStatus);
                }
            });
        }
    }

    static void onAvatarChange(final int error, final String avatarUlr) {
        if (onDeviceLoginListener != null) {
            postMain(new Runnable() {
                @Override
                public void run() {
                    onDeviceLoginListener.onAvatarChange(error, avatarUlr);
                }
            });
        }
    }

    static void onNicNameChange(final int error, final String nickName) {
        if (onDeviceLoginListener != null) {
            postMain(new Runnable() {
                @Override
                public void run() {
                    onDeviceLoginListener.onNicNameChange(error, nickName);
                }
            });
        }
    }

    static void onLogoutComplete(final int error) {
        if (onDeviceLoginListener != null) {
            postMain(new Runnable() {
                @Override
                public void run() {
                    onDeviceLoginListener.onLogoutComplete(error);
                }
            });
        }
    }

    static void onQrCodeScan(final String userName, final String avatar, final int scanTime) {
        Log.d(TAG, "onQrCodeScan, username = " + userName + ", avatar url = " + avatar + ", scan time = " +scanTime);
        if (onDeviceLoginListener != null) {
            postMain(new Runnable() {
                @Override
                public void run() {
                    onDeviceLoginListener.onQrCodeScan(userName, avatar, scanTime);
                }
            });
        }
    }


    /** OnDeviceLoginListener end (绑定者变化的callback在上面)*/


    static void postMain(Runnable runnable) {
        XWSDKJNI.postMain(runnable);
    }


    public static ArrayList<DeviceBinderInfo> getBinderList() {
        return XWSDKJNI.getBinderList();
    }

    public static DeviceLoginInfo getLoginInfo() {
        return DeviceSDK.getInstance().mDeviceLoginInfo;
    }

    /**
     * 根据username查询DeviceContactInfo
     *
     * @param userName 好友的username
     * @return
     */
    public static DeviceBinderInfo getDeviceContactInfo(String userName) {
        DeviceBinderInfo binder = mAllFriendListCache.get(userName);

        return binder;
    }

}
