/*
 * Tencent is pleased to support the open source community by making  XiaoweiSDK Demo Codes available.
 *
 * Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
 *
 * Licensed under the MIT License (the "License"); you may not use this file except in compliance
 * with the License. You may obtain a copy of the License at
 *
 * http://opensource.org/licenses/MIT
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 *
 */
package com.tencent.xiaowei.sdk;

import android.content.Context;
import android.media.AudioManager;
import android.os.Handler;
import android.support.annotation.NonNull;
import android.util.Log;
import com.tencent.xiaowei.def.XWCommonDef;
import com.tencent.xiaowei.info.DeviceBinderInfo;
import com.tencent.xiaowei.info.DeviceLocationInfo;
import com.tencent.xiaowei.info.DeviceLoginInfo;
import com.tencent.xiaowei.info.DeviceOTAInfo;
import com.tencent.xiaowei.info.DeviceProfile;
import com.tencent.xiaowei.info.XWAppInfo;
import com.tencent.xiaowei.info.XWContextInfo;
import com.tencent.xiaowei.info.XWEventLogInfo;
import com.tencent.xiaowei.info.XWPlayStateInfo;
import com.tencent.xiaowei.info.XWRequestInfo;
import com.tencent.xiaowei.info.XWResponseInfo;
import com.tencent.xiaowei.info.XWTTSDataInfo;
import com.tencent.xiaowei.util.Singleton;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

/**
 * 语音服务接口
 */
public class DeviceSDK {
    private static final String TAG = "DeviceSDK";

    private ConcurrentHashMap<String, RequestListener> mMapRequestListener = new ConcurrentHashMap<>();
    private ConcurrentHashMap<String, OnRspListener> mRspListenerMap = new ConcurrentHashMap<>();
    private OnUploadDataListener mUploadDataListener = null;
    private Context mContext;
    private Thread mUiThread;
    private Handler mainHandler;

    private AudioRequestListener audioRequestListener;
    private NetworkDelayListener mNetworkDelayListener;
    private OnReceiveTTSDataListener mOnReceiveTTSDataListener;

    DeviceLoginInfo mDeviceLoginInfo;

    private VoipEventListener mVoipEventListener = null;
    private GpsListener mGpsListener = null;
    private OtaListener mOtaListener = null;
    private Map<String,OtaListener> stringOtaListenerMap = new HashMap<>();

    private DeviceSDK() {

    }

    public String getAssetsCacheFile(Context context,String fileName)   {
        File cacheFile = new File(context.getCacheDir(), fileName);
        try {
            InputStream inputStream = context.getAssets().open(fileName);
            try {
                FileOutputStream outputStream = new FileOutputStream(cacheFile);
                try {
                    byte[] buf = new byte[1024];
                    int len;
                    while ((len = inputStream.read(buf)) > 0) {
                        outputStream.write(buf, 0, len);
                    }
                } finally {
                    outputStream.close();
                }
            } finally {
                inputStream.close();
            }
        } catch (IOException e) {
            e.printStackTrace();
        }
        return cacheFile.getAbsolutePath();
    }

    private static final Singleton<DeviceSDK> sSingleton = new Singleton<DeviceSDK>() {
        @Override
        protected DeviceSDK createInstance() {
            return new DeviceSDK();
        }
    };

    /**
     * 获取语音服务实例
     *
     * @return
     */
    public static DeviceSDK getInstance() {
        return sSingleton.getInstance();
    }


    /**
     * 登录小微设备，所有的服务都应该在设备登录成功后进行
     *
     * @param context   上下文对象，一般传Application的就行
     * @param loginInfo 登录信息，这里一定要全部设置正确，否则SDK可能无法正常工作
     */
    public int login(Context context, DeviceLoginInfo loginInfo, OnDeviceLoginListener listener) {
        if (mContext != null) {
            logout();
        }
        mContext = context.getApplicationContext();

        if (mContext == null) {
            throw new RuntimeException("context cannot be null.");
        }
        if (listener == null) {
            throw new RuntimeException("Init DeviceSDK failed,listener is null.");
        }

        if (mainHandler == null) {
            mainHandler = new Handler(context.getMainLooper());
        }
        if (mUiThread == null) {
            mUiThread = context.getMainLooper().getThread();
        }

        if (loginInfo.serialNumber == null || loginInfo.appid == null || loginInfo.license == null || loginInfo.sysPath == null) {
            return -1;
        }

        mDeviceLoginInfo = loginInfo;
        com.tencent.mars.ilink.comm.PlatformComm.init(mContext, mainHandler);
        XWSDKJNI.getInstance().initJNI(0);// 控制打印native数字越小打印的日志级别越多。
        String caPath = getAssetsCacheFile(mContext,"cacert.pem");

        int ret = XWSDKJNI.getInstance().initDevice(context.getPackageName(), loginInfo.license.getBytes(), loginInfo.serialNumber, loginInfo.productId, loginInfo.productVersion, loginInfo.keyVersion, loginInfo.runMode,
                loginInfo.sysPath, loginInfo.sysCapacity, loginInfo.appid, caPath);

        if (ret == 0) {
            int[] versions = XWSDKJNI.getInstance().getDeviceSDKVersion();
            DeviceBaseManager.setOnDeviceLoginListener(listener);
        }

        return ret;
    }


    /**
     * 获取设备的profile，如设备自己的头像，昵称等等
     * @return
     */
    public DeviceProfile getDeviceProfile() {
        return XWSDKJNI.getDeviceProfile();
    }


    /**
     * 改设备自己的昵称
     *
     * @param nickname 想要改的昵称，不能是空
     * @return 成功或者失败
     */
    public int updateDeviceNickname(String nickname){
        return XWSDKJNI.updateDeviceNickname(nickname);
    }

    /**
     * 初始化小微，一定要在登录成功之后调用
     *
     * @return
     */
    public int startXiaoweiService() {
        return XWSDKJNI.startXiaoweiService();
    }

    /**
     * 改设备自己的头像，请自己确保路径下的图片可用
     * @param jpgFilePath JPG格式的头像图片文件地址，注意是本文件，不是网络文件，512x512 或256x256
     *
     * @return 成功或者失败
     */
    public int updateDeviceAvatar(String jpgFilePath){
        return XWSDKJNI.updateDeviceAvatar(jpgFilePath);
    }

    /**
     * 退出登录，释放资源
     *
     * @return
     */
    public int logout() {
        mMapRequestListener.clear();
        mRspListenerMap.clear();
        mDeviceLoginInfo = null;
        mContext = null;
        mainHandler = null;
        mUiThread = null;

        return XWSDKJNI.getInstance().destroy();
    }

    /**
     * 设备登录相关事件
     */
    public interface OnDeviceLoginListener {
        /**
         * 设备登录成功回调，必须实现，若设备登录不成功，各项服务都无法使用。注意产生这个回调并不意味着上线成功，而是通知上线的结果，需要根据errCode进行判断。
         *
         * @param errorCode 0：上线成功。-1：账号信息有问题。其它错误请联系我们处理。
         */
        void onLoginComplete(int errorCode);

        /**
         * 网络状态变化回调，一般应关注这个回调，当设备掉线后，就不要使用小微的功能了，同时可以播放一些提示，如“网络已断开”
         *
         * @param newStatus 新的网络状态,1：在线，0：离线
         */
        void onNetStatusChange(int newStatus);

        /**
         * 设备自己的昵称被更改的通知，这个更改可能来自SDK，也可能来自app、小程序、甚至后台等，当设备昵称产生变化后，如果是有屏设备，可以更新一下UI显示。设备的默认昵称为厂商在小微后台注册时留下的昵称。
         *
         * @param errorCode 尝试更新昵称失败的原因，除非有特殊配置，或者断网，不然不会失败，如果获得了非0值，请联系我们处理。
         * @param nickName 新的昵称
         */
        void onNicNameChange(int errorCode, String nickName);

        /**
         * 设备头像被更改的通知，类似昵称被更改。
         *
         * @param errorCode
         * @param avatarUrl 新头像的URL
         */
        void onAvatarChange(int errorCode, String avatarUrl);

        /**
         * 退出登录成功，一般用户不需要，除非您需要在程序中反初始化小微，然后以另一套账号信息(SN、licence等)重新登录。
         *
         * @param errorCode 0：退出登录成功。目前不会有其它结论。
         */
        void onLogoutComplete(int errorCode);


        /**
         * 绑定者列表产生变化的通知，登录后会收到绑定者列表，绑定者列表变化也会收到它，主动拉绑定者列表也会收到它。每次回调携带全部的绑定者信息。
         *
         * @param errorCode 无法获取绑定者的原因，若收到了非0值，请联系我们处理。
         * @param arrayList 绑定者列表，第一个绑定者是设备的owner
         */
        void onBinderListChange(int errorCode, ArrayList<DeviceBinderInfo> arrayList);

        /**
         * 好友请求,当某个微信用户通过扫码想绑定设备时，会收到这个回调，收到后可通过SDK提供的处理接口来同意或拒绝。同时在APP和小程序端也会收到这个推送，任意一端都可以处理这个请求。
         *
         * @param info 欲绑定设备的微信用户信息。
         */
        void onBinderVerify(DeviceBinderInfo info);

        /**
         * 动态二维码变了
         *
         * @param qrCodePath 二维码路径
         * @param expireTime 过期绝对时间，秒
         * @param expireIn 过期相对时间，秒
         */
        void onQrCodeRefresh(String qrCodePath, int expireTime, int expireIn);

        /**
         * 二维码被人扫的通知
         * @param username 扫码者微信昵称
         * @param avatarUrl 扫码者微信头像URL
         * @param scanTime 扫码时间
         */
        void onQrCodeScan(String username, String avatarUrl, int scanTime);
    }

    /**
     * 设置通用的请求状态回调接口，十分重要的接口，小微的大部分response都会从这个接口中返回，参考官网文档说明
     *
     * @param listener
     */
    public void setAudioRequestListener(AudioRequestListener listener) {
        this.audioRequestListener = listener;
    }

    /**
     * 发起语音或文本请求，具体参考官网文档说明，在语音请求时，连续多次调用此接口来fill audio data
     * @param type {@link XWCommonDef.RequestType}
     * @param requestData 请求数据，具体定义参考官网文档
     * @param param 请求参数
     * @return 返回String为voiceId，即请求Id，如果是空则表示请求失败
     */
    public String request(int type, byte[] requestData, XWRequestInfo param) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        String strVoiceID = XWSDKJNI.request(type, requestData, param);
        if (strVoiceID.isEmpty()) {
            Log.e(TAG, "request voiceID is null.");
            return null;
        }

        return strVoiceID;
    }

    /**
     * 发起语音或文本请求，并使用默认的上下文参数，主要用于文本请求
     * @param type {@link XWCommonDef.RequestType}
     * @param requestData 请求数据，具体定义参考官网文档
     * @return 返回String为voiceId，即请求Id，如果是空则表示请求失败
     */
    public String request(int type, byte[] requestData) {
        return request(type, requestData, (XWRequestInfo) null);
    }


    /**
     * 语音请求  use {@link #request} instead.
     *
     * @param type        请求类型 {@link XWCommonDef.RequestType}
     * @param requestData 请求数据
     * @param context     上下文，用于携带额外的会话信息
     * @return 本次请求对应的voiceID
     */
    @Deprecated
    public String request(int type, byte[] requestData, XWContextInfo context) {
        return request(type, requestData, context.toRequestInfo());
    }


    /**
     * 取消语音请求
     *
     * @param voiceId 要取消的voiceID，当为null的时候，表示取消所有请求
     */
    public int requestCancel(String voiceId) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.cancelRequest(voiceId);
    }

    /**
     * @Deprecated 根据文本转TTS，use{@linke #request(String cmd, String subCmd, String params, OnRspListener listener)} instead
     *
     * @param strText  请求的文本
     * @param listener 回调监听
     * @return 本次请求对应的VoiceID
     */
    @Deprecated
    public String requestTTS(@NonNull byte[] strText, RequestListener listener) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        String strVoiceID = XWSDKJNI.request(XWCommonDef.RequestType.ONLY_TTS, strText, null);

        if (strVoiceID.isEmpty()) {
            Log.e(TAG, "request voiceID is null.");
            return strVoiceID;
        }

        if (listener != null) {
            mMapRequestListener.put(strVoiceID, listener);
        }

        return strVoiceID;
    }

    /**
     * 取消TTS的传输，目前TTS不再是opus流了，理论上这个接口已经不需要再使用了
     *
     * @param resId
     */
    @Deprecated
    public void cancelTTS(String resId) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        XWSDKJNI.cancelTTS(resId);
    }

    /**
     * 设置QQ音乐品质，这里其实就是包装了一下requestCmd，建议直接使用它
     *
     * @param quality {@link XWCommonDef.PlayQuality}
     * @return 返回值请参考 {@link XWCommonDef.ErrorCode}
     */
    @Deprecated
    public int setMusicQuality(int quality) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        String cmd = "{\"skill_info\":{\"id\":\"8dab4796-fa37-4114-0011-7637fa2b0001\",\"name\":\"音乐\"},\"quality_config\":{\"type\":0,\"quality\":" + quality + "}}";
        XWSDKJNI.requestCmd("PLAY_RESOURCE", "quality_config", cmd);
        return 0;
    }

    /**
     * 切换线程用，外部不应调用
     * @param runnable
     */
    void runOnMainThread(Runnable runnable) {
        if (mUiThread == Thread.currentThread()) {
            runnable.run();
            return;
        }

        if (mainHandler != null) {
            mainHandler.post(runnable);
        }
    }

    /**
     * JNI上来的回调
     * @param voiceID
     * @param event
     * @param rspData
     * @param extendData
     * @return
     */
    boolean onRequest(final String voiceID, final int event, final XWResponseInfo rspData, final byte[] extendData) {
        RequestListener listener = mMapRequestListener.remove(voiceID);
        if (listener != null) {
            return listener.onRequest(event, rspData, extendData);
        } else if (audioRequestListener != null) {
            return audioRequestListener.onRequest(voiceID, event, rspData, extendData);
        }

        return false;
    }

    @Deprecated
    boolean onTTSPushData(String voiceId, XWTTSDataInfo tts) {
        if (mOnReceiveTTSDataListener != null) {
            return mOnReceiveTTSDataListener.onReceive(voiceId, tts);
        }
        return false;
    }

    long lastNetDelayTime;
    boolean onNetWorkDelay(final String voiceID, final long time) {
        if (mNetworkDelayListener != null) {
            runOnMainThread(new Runnable() {
                @Override
                public void run() {
                    mNetworkDelayListener.onDelay(voiceID, time);
                }
            });
        }
        lastNetDelayTime = time;
        return true;
    }

    /**
     * 获取系统媒体音量用于上报，
     * @return
     */
    public int getVolume() {
        int tag = AudioManager.STREAM_MUSIC;
        AudioManager mAudioManager = (AudioManager) mContext.getSystemService(Context.AUDIO_SERVICE);
        if (mAudioManager.getMode() == AudioManager.MODE_IN_COMMUNICATION) {
            tag = AudioManager.MODE_IN_COMMUNICATION;
        }
        int max = mAudioManager.getStreamMaxVolume(tag);
        int current = mAudioManager.getStreamVolume(tag);
        int vol = (int) (current * 100.0f / max);
        Log.d(TAG, "getVolume current: " + current + ", max: " + max + ", vol: " + vol);
        return vol;
    }

    /**
     * 上报播放状态
     *
     * @param stateInfo 要上报的当前状态
     */
    public int reportPlayState(XWPlayStateInfo stateInfo) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        stateInfo.volume = getVolume();
        Log.d(TAG, "reportPlayState " + stateInfo);
        return XWSDKJNI.reportPlayState(stateInfo);
    }

    /**
     * 上报音量
     *
     */
    public int reportVolumeChange() {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.reportVolumeChange(getVolume());
    }


    /**
     * 通用请求回调通知
     */
    public interface RequestListener {
        /**
         * 收到通知
         * 根据event分为多种事件
         *
         * @param event      当前通知事件 {@link XWCommonDef.XWEvent}
         * @param rspData    标准信息
         * @param extendData 扩展信息
         */
        boolean onRequest(int event, XWResponseInfo rspData, byte[] extendData);
    }

    /**
     * 语音服务状态回调接口
     */
    public interface AudioRequestListener {
        /**
         * 收到通知
         * 根据event分为多种事件
         *
         * @param voiceId    当次请求的VoiceId
         * @param event      当前通知事件 {@link XWCommonDef.XWEvent}
         * @param rspData    标准信息
         * @param extendData 扩展信息
         * @return 收到状态通知后，上层是否需要处理
         */
        boolean onRequest(String voiceId, int event, XWResponseInfo rspData, byte[] extendData);
    }

    /**
     * 网络延迟的监听，每次小微的网络请求都会回调，从开始发送到收到响应的时差
     */
    public interface NetworkDelayListener {
        void onDelay(String voiceId, long ms);
    }

    /**
     * 设置网络延迟的监听
     *
     * @param listener {@link NetworkDelayListener}
     */
    public void setNetworkDelayListener(NetworkDelayListener listener) {
        mNetworkDelayListener = listener;
    }

    /**
     * 网络延迟的监听，每次小微的网络请求都会回调，从开始发送到收到响应的时差
     */
    public interface OnReceiveTTSDataListener {
        boolean onReceive(String voiceId, XWTTSDataInfo ttsData);
    }

    /**
     * 设置接受TTS数据的监听
     *
     * @param listener {@link OnReceiveTTSDataListener}
     */
    public void setOnReceiveTTSDataListener(OnReceiveTTSDataListener listener) {
        mOnReceiveTTSDataListener = listener;
    }

    /**
     * 上报日志文件
     */
    public void uploadLogs(String start, String end) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        XWSDKJNI.getInstance().upLoadLog();
    }

    /**
     * 设置设备场景
     * @param sceneType 场景类型，可以一次设置多个场景，可以为空数组
     * @param sceneInfo 场景类型对应的信息，与场景一一对应，其中的每个元素不得是null
     * @return
     */
    public int setDeviceScene(int[] sceneType, String[] sceneInfo) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        if (sceneType != null && sceneInfo != null) {
            if (sceneType.length != sceneInfo.length) {
                return -1;
            }
            for (int i=0; i< sceneInfo.length; ++i) {
                if (sceneInfo[i] == null) {
                    return -1;
                }
            }
        }
        else if (sceneType != null || sceneInfo != null) {
            return -1;
        }
        return XWSDKJNI.getInstance().setDeviceScene(sceneType, sceneInfo);
    }

    /**
     * 设置设备的可用状态
     * @param av
     * @return
     */
    public int setDeviceAvailability(int av) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.setDeviceAvailability(av);
    }
    /**
     * 设置声波配网的cacheKey
     * @return
     */
    public int setCacheKey(String k, String v) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.getInstance().setCacheKey(k,v);
    }

    /**
     * 用户不满意上次的识别，voiceID传入，上报后台
     */
    public void errorFeedBack(String voiceID) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        if(voiceID == null )
        {
            return;
        }
        XWSDKJNI.errorFeedBack(voiceID);
    }

    /**
     * 重置系统时钟
     * @return
     */
    public int manulSetClock() {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.manulSetClock();
    }

    /**
     * 查询设备睡眠模式
     * @return 0正常，1睡眠模式
     */
    public int getSleepState() {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.getSleepState();
    }

    /**
     * 临时屏蔽5分钟睡眠模式
     * @return
     */
    public int maskSleepModeTemporarily() {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.maskSleepModeTemporarily();
    }


    /**
     * 上传本地信息的结果回调通知
     */
    public interface OnUploadDataListener {
        /**
         * 回调
         *
         * @param voiceId 请求的voiceid
         * @param errCode 0:成功 非0:失败
         */
        void OnReply(String voiceId, int errCode);
    }

    public void setUploadDataListener(OnUploadDataListener in) {
        mUploadDataListener = in;
    }

    /**
     * JNI层上来的回调
     * @param errCode
     */
    void onUploadData(String voiceId, int errCode) {
        Log.v(TAG, "onUploadDataRet errCode:" + errCode);
        if (mUploadDataListener != null) {
            mUploadDataListener.OnReply(voiceId, errCode);
        }
    }

    /**
     * 上传本地数据到云端，用于一些本地相关技能
     * @param type {@link XWCommonDef.DataListType}
     * @param data 参考官网文档的对应说明
     * @return voiceId
     */
    public String uploadData(int type, String[] data) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.uploadDatalist(type, data);
    }

    /**
     * 开启可见可答，开启后，每次请求将携带词表。使用{@link #setWordslist}来设置
     *
     * @param enable true:open, false:close
     * @return 0:success else failed
     */
    public int enableV2A(boolean enable) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.enableV2A(enable);
    }

    /**
     * 设置词库列表，设置后一直有效，调用enableV2A后选择是否携带此词表
     *
     * @param type       词库类型 {@link XWCommonDef.WordListType}
     * @param words_list 词库
     * @return 0:success else failed
     */
    public int setWordslist(int type, String[] words_list) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.setWordslist(type, words_list);
    }

    /**
     * 收藏FM或者QQ音乐，本质上就是包装了一下requestCmd，
     *
     * @param appInfo    场景信息
     * @param playId     资源ID
     * @param isFavorite true表示收藏，false表示取消收藏
     */
    public void setFavorite(XWAppInfo appInfo, String playId, boolean isFavorite) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }

        String cmd = "{\"skill_info\":{\"id\":\"" + appInfo.ID + "\",\"name\":\"音乐\"},\"favorite\":" + isFavorite + ",\"play_id\":\"" + playId + "\"}";

        XWSDKJNI.requestCmd("PLAY_RESOURCE", "set_favorite", cmd);
    }

    /**
     * 上报用户自己的流水日志，一般有重大问题再调用
     * @param log
     */
    public void reportUserLog(XWEventLogInfo log) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        XWSDKJNI.reportUserLog(log);
    }

    /**
     * 通用请求资源的回调
     */
    public interface OnRspListener {
        /**
         * @param voiceId 请求唯一标识
         * @param error   错误码
         * @param json    结果
         */
        void onRsp(String voiceId, int error, String json);
    }

    /**
     * 通用请求资源的接口，参考官网文档的通用请求
     *
     * @param cmd      主命令
     * @param subCmd   子命令
     * @param params   参数
     * @param listener {@link OnRspListener}
     */
    public String request(String cmd, String subCmd, String params, OnRspListener listener) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        String strVoiceID = XWSDKJNI.requestCmd(cmd, subCmd, params);
        if (!strVoiceID.isEmpty() && listener != null) {
            Log.d(TAG, "request voiceId: " + strVoiceID);
            mRspListenerMap.put(strVoiceID, listener);
        }
        return strVoiceID;
    }

    /**
     * JNI层上来的包装
     * @param voiceId
     * @param error
     * @param json
     */
    public void onCommonRequest(final String voiceId, final int error, final String json) {
        final OnRspListener listener = mRspListenerMap.get(voiceId);
        if (listener != null) {
            mainHandler.post(new Runnable() {
                @Override
                public void run() {
                    listener.onRsp(voiceId, error, json);
                }
            });
        }
    }

    /**
     * 设置GPS回调，如果设备支持GPS的话，设置了这个回调之后，小微SDK会在需要的时候来回调这个方法来获取GPS信息
     * @param listener
     */
    public void setGpsCallback(GpsListener listener){
        mGpsListener = listener;
        XWSDKJNI.setGpsCallback();
    }

    /**
     * GPS请求的JNI回调，注意这个方法不能阻塞，一定要快，所以在这里面一般应该是去读一下缓存
     */
    public DeviceLocationInfo onGpsJniCallback() {
        if (mGpsListener != null) {
            return mGpsListener.onGpsCallback();
        }
        return new DeviceLocationInfo(false, 0, 0, 0, 0, 0);
    }

    /**
     * GPS请求的interface
     */
    public interface GpsListener {
        DeviceLocationInfo onGpsCallback();
    }

    /**
     * 打开voip服务，如果不开就无法收到voip通知，这里的参数设置错误可能导致VOIP无法使用或功能缺陷。
     * @param audioType 0 silk，1 silk+aac-ld, only silk+aac-ld support
     * @param closeAV 1:close audio，2:close video, 0:default
     * @param videoRatio 0:4/3，1:16/9，2:1/1，3:1/2
     * @param defaultVideoLength default length for long side, such as 640.(e.g. if video_ratio = 0, it will be 640:480)
     * @param cpuCoreNum CPU core number, must > 0
     * @param cpuFreq CPU frequency（unit: MHz）
     * @param cpuArch CPU architecture（1: armv5，2: armv6，4: armv7）
     * @return 非0则失败
     */
    public int startVoipService(
        int audioType, int closeAV, int videoRatio, int defaultVideoLength, int cpuCoreNum, int cpuFreq, int cpuArch) {
        if (mContext == null) {
            throw new RuntimeException("You need to call login at first.");
        }
        return XWSDKJNI.startVoipService(audioType, closeAV, videoRatio, defaultVideoLength, cpuCoreNum, cpuFreq, cpuArch);
    }

    /**
     * 挂断电话，用于接听时的拒绝或者通话过程中的主动挂断
     */
    public int hangUpVoipCall() {
        Log.e("VOIP","HangupVoip by user");
        return XWSDKJNI.hangUpVoipCall();
    }

    /**
     * 收到来电时接电话，如果要拒绝，则调用hangUpVoipCall()
     */
    public int answerVoipCall() {
        return XWSDKJNI.answerVoipCall();
    }

    /**
     * 发起voip请求
     * @param username 打给谁
     * @param appid 打给哪个小程序，不设置就默认打给小微小程序
     * @param type 1：音视频通话 2：纯音频
     */
    public int createVoipCall(String username, String appid, int type) {
        return XWSDKJNI.createVoipCall(username, appid, type);
    }

    /**
     * 发送音频数据
     */
    public int SendAudioData(byte[] buffer, int len, int delay) {
        return XWSDKJNI.SendAudioData(buffer, len, delay);
    }

    /**
     * 发送视频数据
     */
    public int SendVideoData(byte[] img, int len, int w, int h, int f) {
        return XWSDKJNI.SendVideoData(img, len, w, h, f);
    }

    /**
     * 获取音频数据
     */
    public int GetAudioData(byte[] playAudio, int len) {
        return XWSDKJNI.GetAudioData(playAudio, len);
    }

    /**
     * 获取视频信息
     */
    public XWSDKJNI.VideoInfo getVideoInfo() {
        return XWSDKJNI.getVideoInfo();
    }

    /**
     * 获取视频数据
     */
    public int GetDecodeVideoData(byte [] iImg) {
        return XWSDKJNI.GetDecodeVideoData(iImg);
    }
    
    /**
     * 切换自己的音视频是否打开，默认都是打开的，例如关掉了自己的麦克风，就isOpenAudio = false
     * @param isOpenAudio 自己是否打开了音频
     * @param isOpenVideo 自己是否打开了视频
     * @return
     */ 
    public int switchVoipAV(boolean isOpenAudio, boolean isOpenVideo) {
        return XWSDKJNI.switchVoipAV(isOpenAudio, isOpenVideo);
    }

     /**
      * 获取某个人是否正在说话，返回1就是在说话，0未说话，其它查无此人
      * @param memberId 注意voip这边管理都是用memberId的，它对应的openid需要在onVoipMemberChange(...)中获得
      * @return
      */
    public int getVoipVoiceActivity(int memberId) {
        return XWSDKJNI.getVoipVoiceActivity(memberId);
    }

    /**
     * 动态设置voip参数
     */
    public int setVoipAppCmd(int type, byte[] buffer, int len)
    {
        return XWSDKJNI.setVoipAppCmd(type, buffer, len);
    }

    /**
     * 硬编前处理
     */
    public  int voipVideoPreEncProcessCloud(byte[] recordData,int len,int videoWidth,int videoHeight,int format,byte[] encImg)
    {
        return XWSDKJNI.voipVideoPreEncProcessCloud(recordData, len, videoWidth, videoHeight, format, encImg);
    }


    public XWSDKJNI.HWEncInfo getHWEncInfo() {
        return XWSDKJNI.getHWEncInfo();
    }


    public int voipSetVideoResolution(byte[] pbBuf, int len) {
        return XWSDKJNI.voipSetVideoResolution(pbBuf, len);
    }

    /**
     * 获取关闭麦克风的成员列表
     * @return 关闭麦克风的成员列表，里面是memberId，注意返回可能是空数组
     */ 
    public int[] getVoipVadMember() {
        return XWSDKJNI.getVoipVadMember();
    }

    /**
     * 设置VOIP和微信消息的监听
     */
    public void setVoipEventListener(VoipEventListener listener) {
        this.mVoipEventListener = listener;
    }

    /**
     * VOIP和微信消息通用listener
     */
    public interface VoipEventListener {
        /**
         * 收到voip请求（app或者小程序打给设备）
         * @param fromUsername 发起者的ID
         */
        void onVoipInvite(String fromUsername, int type);
        /**
         * 加入房间结果
         * @param result 结果，0成功，非0失败
         * @param isMaster 是否主叫，true主叫，false被叫
         */
        void onJoinRoom(int result, boolean isMaster);
        /**
         * 主叫的时候邀请对方结果
         * @param result 结果，0成功，非0失败
         */
        void onInviteResult(int result);
        /**
         * voip取消，包括主被叫时对方挂断，以及被叫时加入一个非法的房间
         * @param isMaster 是否主叫，true主叫，false被叫
         */
        void onVoipCancel(boolean isMaster);
        /**
         * voip接通
         */
        void onVoipTalking();
        /**
         * voip结束
         */
        void onVoipFinish();
        /**
         * voip 参数发送变化，客户端应该严格按照这个参数来配置
         * @param sampleRate 音频数据采样率
         * @param sampleLenInms 每一帧的时间长度
         * @param extendPB 额外的PB信息，先不管
         */
        void onVoipParamChange(int sampleRate, int sampleLenInms, String extendPB);

        /**
         * voip房间人数变化发生的通知
         * @param members 目前房间内的数数组，注意可能是空数组，数组内容是memberid
         * @param usernames 房间内的人，每个id对应的username
         */
        void onVoipMemberChange(int[] members, String[] usernames);
        
        /**
         * voip房间内的用户关闭或者打开麦克风、摄像头的回调，一旦发生变化就会回调。这里始终展示当前打开麦克风和打开摄像头的用户
         * @param audioMembers 打开麦克风的人的ID列表，注意可能是空数组
         * @param videoMembers 打开视频的人的ID列表，注意可能是空数组
         */
        void onVoipAVMemberChange(int[] audioMembers, int[] videoMembers);
    }

    public void onVoipInvite(String fromUsername, int type)
    {
        Log.v(TAG, "onVoipInvite fromUsername:" + fromUsername + " ,type = " + type);
        if (mVoipEventListener != null) {
            mVoipEventListener.onVoipInvite(fromUsername, type);
        }
    }

    public void onJoinRoom(int result, boolean isMaster)
    {
        Log.v(TAG, "onJoinRoom isMaster:" + isMaster);
        if (mVoipEventListener != null) {
            mVoipEventListener.onJoinRoom(result, isMaster);
        }
    }

    public void onInviteResult(int result)
    {
        Log.v(TAG, "onInviteResult result:" + result);
        if (mVoipEventListener != null) {
            mVoipEventListener.onInviteResult(result);
        }
    }

    public void onVoipCancel(boolean isMaster)
    {
        Log.v(TAG, "onVoipCancel isMaster:" + isMaster);
        if (mVoipEventListener != null) {
            mVoipEventListener.onVoipCancel(isMaster);
        }
    }

    public void onVoipTalking()
    {
        Log.v(TAG, "onVoipTalking");
        if (mVoipEventListener != null) {
            mVoipEventListener.onVoipTalking();
        }
    }

    public void onVoipFinish()
    {
        Log.v(TAG, "onVoipFinish");
        if (mVoipEventListener != null) {
            mVoipEventListener.onVoipFinish();
        }
    }

    public void onVoipParamChange(int sampleRate, int sampleLenInms, String extendPB)
    {
        Log.v(TAG, "onVoipParamChange, sampleLenInms = " + sampleLenInms);
        if (mVoipEventListener != null) {
            mVoipEventListener.onVoipParamChange(sampleRate,sampleLenInms,extendPB);
        }
    }

    public void onVoipMemberChange(int[] members, String[] usernames)
    {
        Log.v(TAG, "onVoipMemberChange");
        if (members != null) {
            for (int i=0;i<members.length;i++) {
                Log.d(TAG, "memberid:" + members[i] + ". username = " +usernames[i]);
            }
        }
        if (mVoipEventListener != null) {
            mVoipEventListener.onVoipMemberChange(members, usernames);
        }
    }

    public void onVoipAVMemberChange(int[] audioMembers, int[] videoMembers)
    {
        Log.v(TAG, "onVoipAVMemberChange");
        if (audioMembers != null) {
            for (int i=0;i<audioMembers.length;i++) {
                Log.d(TAG, "audio memberid:" + audioMembers[i] );
            }
        }
        if (videoMembers != null) {
            for (int i=0;i<videoMembers.length;i++) {
                Log.d(TAG, "video memberid:" + videoMembers[i]);
            }
        }
        if (mVoipEventListener != null) {
            mVoipEventListener.onVoipAVMemberChange(audioMembers, videoMembers);
        }
    }

    //voip end

    //ota
    /**
     *OTA listener
     */
    public interface OtaListener {
        /**
         * 收到OTA检查结果
         */
        void onOtaCheckResult(int errCode, String businessID, boolean newVersion, DeviceOTAInfo otaInfo);
        /**
         * 下载进度
         */
        void onOtaDownLoadProgress(String url, double current, double total);
        /**
         * 下载成功
         */
        void onOtaDownLoadError(String url, int errCode);
        /**
         * 下载失败
         */
        void onOtaDownLoadComplete(String url, String filePath);
    }

    public void setOtaListener(String url,OtaListener listener) {
        if (url.equals("")) { //ota下载
            mOtaListener = listener;
        } else { //其他资源下载
            stringOtaListenerMap.put(url, listener);
        }
    }

    public void onOtaCheckResult(int errCode, String businessID, boolean newVersion, DeviceOTAInfo otaInfo) {
        Log.v(TAG, "onOtaCheckResult errCode:" + errCode + ",businessID:" + businessID + ",newVersion:" + newVersion + ",otaInfo:" + otaInfo);
        if (mOtaListener != null) {
            mOtaListener.onOtaCheckResult(errCode, businessID, newVersion, otaInfo);
        }
    }

    public void onOtaDownLoadProgress(String url, double current, double total) {
        Log.v(TAG, "onOtaDownLoadProgress url:" + url + ",current:" + current + ",total:" + total);
        if (stringOtaListenerMap.get(url) != null) {
            stringOtaListenerMap.get(url).onOtaDownLoadProgress(url, current, total);
        } else if (mOtaListener != null) {
            mOtaListener.onOtaDownLoadProgress(url, current, total);
        }

    }

    public void onOtaDownLoadError(String url, int errCode) {
        Log.v(TAG, "onOtaDownLoadError url:" + url + ",errCode:" + errCode);
        if (stringOtaListenerMap.get(url) != null) {
            stringOtaListenerMap.get(url).onOtaDownLoadError(url, errCode);
            stringOtaListenerMap.remove(url);
        } else if (mOtaListener != null) {
            mOtaListener.onOtaDownLoadError(url, errCode);
        }
    }

    public void onOtaDownLoadComplete(String url, String filePath) {
        Log.v(TAG, "onOtaDownLoadComplete url:" + url + ",filePath:" + filePath);
        if (stringOtaListenerMap.get(url) != null) {
            stringOtaListenerMap.get(url).onOtaDownLoadComplete(url, filePath);
            stringOtaListenerMap.remove(url);
        } else if (mOtaListener != null) {
            mOtaListener.onOtaDownLoadComplete(url, filePath);
        }

    }

    /**
     *cloud ticket listener
     */
    public interface CloudTicketListener {
        /**
         *获取云端ticket的callback
         */
        void onGetCloudTicket(String ticket,int expire_seconds);
    }

    public int GetCloudTicket(CloudTicketListener listener) {
        return XWSDKJNI.getCloudTicket(listener);
    }
    public int deviceConfigOta() {
        return XWSDKJNI.deviceConfigOta();
    }

    public int queryOtaUpdate(String businessID, int channel, int otaType, int debugFlag) {
        return XWSDKJNI.queryOtaUpdate(businessID,channel,otaType,debugFlag);
    }

    public int otaDwonload(String url, String downloadPath, int hashType, String hashValue) {
        return XWSDKJNI.otaDwonload(url,downloadPath,hashType,hashValue);
    }

    public int stopOtaDownload(String url) {
        return XWSDKJNI.stopOtaDownload(url);
    }

    public String visitLocalClocks() {
        return XWSDKJNI.visitLocalClocks();
    }

    /**
     * 创建解码工具
     * @return
     */
    public long decoderCreate(String key ) {
        return XWSDKJNI.decoderCreate(key);
    }

    /**
     * 开始解码
     * @return
     */
    public void decoderDecode(long decoder , byte[] data,int size ,int offset) {
        XWSDKJNI.decoderDecode(decoder, data, size, offset);
    }

    /**
     * 销毁解码工具
     * @return
     */
    public void decoderDestroy(long decoder ) {
        XWSDKJNI.decoderDestroy(decoder);
    }
}
